#ifndef RENDERERPROGRAMMABLE_H
#define RENDERERPROGRAMMABLE_H

#include "renderer.h"

#include <glm/glm.hpp>
#include <glm/gtc/matrix_projection.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/transform2.hpp>
#include <glm/gtx/inverse_transpose.hpp>

#include <stack>

using namespace glm;

#define STRINGFY(X) #X

class RendererProgrammable : public Renderer
{
	public:

		RendererProgrammable(void) : angle(0.0f), vertexShader(0), fragmentShader(0), program(0) { }

		virtual void initialize(void)
		{
			const char * vertexShaderSource = STRINGFY
			(
				#version 140\n

				uniform mat4 worldViewProjectionMatrix;
				uniform mat3 viewSpaceNormalMatrix;

				in  vec3 a_position;
				in  vec3 a_normal;

				out vec3 v_normal;

				void main(void)
				{
					v_normal    = viewSpaceNormalMatrix     * a_normal;
					gl_Position = worldViewProjectionMatrix * vec4(a_position, 1.0);
				}
			);
			this->vertexShader = glCreateShader(GL_VERTEX_SHADER);
			glShaderSource(this->vertexShader, 1, &vertexShaderSource, 0);
			glCompileShader(this->vertexShader);

			const char * fragmentShaderSource = STRINGFY
			(
				#version 140\n

				uniform vec4 light0Position;

				in  vec3 v_normal;

				out vec4 o_color;

				void main(void)
				{
					vec3  normal  = normalize(v_normal);
					float lambert = dot(normal, light0Position.xyz);
					o_color       = vec4(vec3(lambert), 1.0);
				}
			);
			this->fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
			glShaderSource(this->fragmentShader, 1, &fragmentShaderSource, 0);
			glCompileShader(this->fragmentShader);

			this->program = glCreateProgram();
			glAttachShader(this->program, this->vertexShader);
			glAttachShader(this->program, this->fragmentShader);

			glBindAttribLocation(this->program, 0, "a_position");
			glBindAttribLocation(this->program, 1, "a_normal");

			glBindFragDataLocationEXT(this->program, 0, "o_color");

			glLinkProgram(this->program);

			/*
			GLint ls = 0;
			glGetProgramiv(this->program, GL_LINK_STATUS, &ls);
			printf("%d\n", ls);

			GLint len = 0;
			glGetProgramiv(this->program, GL_INFO_LOG_LENGTH, &len);
			if (len > 0)
			{
				char * cl = new char[len + 1];
				glGetProgramInfoLog(this->program, len, &len, cl);
				if (len > 0)
				{
					cl[len-1] = '\0';
					printf("%s\n", cl);
				}
				delete [] cl;
			}
			*/
		}

		virtual void finalize(void)
		{
			glDeleteProgram(this->program);
			this->program = 0;

			glDeleteShader(this->vertexShader);
			this->vertexShader = 0;

			glDeleteShader(this->fragmentShader);
			this->fragmentShader = 0;
		}

		virtual void update(float dt)
		{
			this->angle += 90.0f * dt;
		}

		virtual void beginRender(int width, int height)
		{
			// setup projection matrix
			this->projectionMatrix = glm::perspective(60.0f, float(width) / float(height), 0.1f, 10.0f);

			// setup view matrix
			this->viewMatrix = glm::lookAt(vec3(0.0f, 0.0f, 2.0f), vec3(0.0f, 0.0f, 0.0f), vec3(0.0f, 1.0f, 0.0f));

			glEnable(GL_DEPTH_TEST);

			glUseProgram(this->program);

			const GLfloat sceneAmbient[4] = { 0.2f, 0.2f, 0.2f, 1.0f };
			this->setUniform4f("sceneAmbient", sceneAmbient);

			{
				// specify light position with identity modelview matrix (make a "headlight")
				const GLfloat light0Position[4] = { 0.0f, 0.0f, 1.0f, 0.0f };
				this->setUniform4f("light0Position", light0Position);

				const GLfloat light0Ambient[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
				this->setUniform4f("light0Ambient", light0Ambient);

				const GLfloat light0Diffuse[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
				this->setUniform4f("light0Diffuse", light0Diffuse);

				const GLfloat light0Specular[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
				this->setUniform4f("light0Specular", light0Specular);
			}

			this->worldStack.push(glm::rotate(this->angle, vec3(0.0f, 1.0f, 0.0f)));
		}

		virtual void render(const Mesh & m)
		{
			// setup world (model, object) matrix
			this->worldStack.push(this->worldStack.top() * glm::translate(vec3(m.position[0], m.position[1], m.position[2])));

			// setup material
			{
				const float matEmission[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
				this->setUniform4f("matEmission", matEmission);

				//const float matAmbient[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
				const float * matAmbient = m.color;
				this->setUniform4f("matAmbient", matAmbient);

				//const float matDiffuse[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
				const float * matDiffuse = m.color;
				this->setUniform4f("matDiffuse", matDiffuse);

				const float matSpecular[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
				this->setUniform4f("matSpecular", matSpecular);

				const float matShininess[1] = { 0.0f };
				this->setUniform4f("matShininess", matShininess);
			}

			const mat4 worldViewMatrix           = this->viewMatrix * this->worldStack.top();
			const mat4 worldViewProjectionMatrix = this->projectionMatrix * worldViewMatrix;
			const mat3 viewSpaceNormalMatrix     = glm::inverseTranspose(glm::mat3(worldViewMatrix));

			this->setUniform4x4f("worldViewProjectionMatrix", &(worldViewProjectionMatrix[0][0]));
			this->setUniform3x3f("viewSpaceNormalMatrix",     &(viewSpaceNormalMatrix[0][0]));

			m.render();

			this->worldStack.pop();
		}

		virtual void endRender(void)
		{
			this->worldStack.pop();

			glUseProgram(0);

			glDisable(GL_DEPTH_TEST);
		}

	protected:

		float  angle;

		GLuint vertexShader;
		GLuint fragmentShader;
		GLuint program;

		mat4 projectionMatrix;
		mat4 viewMatrix;

		std::stack<mat4> worldStack;

		void setUniform1f(const char * name, const GLfloat value)
		{
			const GLint loc = glGetUniformLocation(this->program, name);
			if (loc < 0) return;
			glUniform1f(loc, value);
		}

		void setUniform4f(const char * name, const GLfloat * values)
		{
			const GLint loc = glGetUniformLocation(this->program, name);
			if (loc < 0) return;
			glUniform4f(loc, values[0], values[1], values[2], values[3]);
		}

		void setUniform3x3f(const char * name, const GLfloat * values)
		{
			const GLint loc = glGetUniformLocation(this->program, name);
			if (loc < 0) return;
			glUniformMatrix3fv(loc, 1, GL_FALSE, values);
		}

		void setUniform4x4f(const char * name, const GLfloat * values)
		{
			const GLint loc = glGetUniformLocation(this->program, name);
			if (loc < 0) return;
			glUniformMatrix4fv(loc, 1, GL_FALSE, values);
		}
};

#undef STRINGFY

#endif // RENDERERPROGRAMMABLE_H
